@@ -9,10 +9,10 @@ require Exporter;
 %EXPORT_TAGS = (':all' => [@EXPORT_OK, @EXPORT]);
 @ISA = qw(Exporter);
 
-$VERSION = 0.02;
+$VERSION = 0.03;
 
 sub BOTTOM () {			# the universal bad value
-    if ($::DEBUG =~ /b/) {
+    if ($::FP_DEBUG =~ /b/) {
 	use Carp 'confess';
 	confess("Bottom!");
     }
@@ -30,6 +30,9 @@ sub numeric {			# check for 2 integer args
 ######################################################################
 ## Parser
 
+##############################
+# Debugging
+
 sub info {			# pretty debugging output.
     my ($pack, $fn, $line, $subr) = caller 1;
     $subr =~ s/^.*:://;
@@ -37,15 +40,21 @@ sub info {			# pretty debugging output.
 }
 
 sub Dparse {			# parse-time debugging output
-    goto &info if $::DEBUG =~ /p/;
+    goto &info if $::FP_DEBUG =~ /p/;
 }
 
 sub Drun {			# run-time debugging.
-    goto &info if $::DEBUG =~ /r/;
+    goto &info if $::FP_DEBUG =~ /r/;
 }
 
+##############################
+# Utilities
+
 # XXX: this shouldn't be needed.  It makes <X> behave the same as X
-# when passed as an argument list.
+# when passed as an argument list.  Single-element lists and scalar
+# values aren't the same kind of thing, but we're trying to pretend as
+# if they are.  Otherwise, perl functions called from FP will all have
+# to take array-refs.
 sub as_array($) {
     my $a = shift;
     if (ref $a eq 'ARRAY') {
@@ -56,7 +65,7 @@ sub as_array($) {
 }
 
 # XXX: this is the disgusting inverse of as_array
-sub to_array {
+sub to_arrayref {
     if (@_ == 1) {
 	return shift;
     } else {
@@ -64,245 +73,152 @@ sub to_array {
     }
 }
 
-sub call_it {			# look up a function
+sub call_it {			# call a coderef, with verbosity
     my $f = shift;
     Drun "Calling $f (@_)";
-    $f->(@_);
+    my @res = $f->(@_);
+    Drun "-> (@res)";
+    @res;
 }
 
-sub findsym {		# look up a function
+sub term {			# create a typed parse-tree node.
+    my $type = shift;
+    return { type => $type, val => [@_] };
+}
+
+##############################
+# Symbol lookup.
+
+# Note: we need to do a bit of magic here to look up functions and
+# variables in both Language::FP and the calling package.
+
+sub findsym {			# look up a function
     my ($sym, $type) = @_;
-    return $sym if ref $sym eq $type;
-    return *{__PACKAGE__.'::'.$sym}{$type} || *{pkg().'::'.$sym}{$type}
-	|| undef;
-}
-
-sub do_bu {			# bu (i.e. currying)
-    my ($f, $o) = @_;
-    Dparse "using $f($o, ...)";
-    return sub {
-	no strict 'refs';
-	Drun "bu $f ($o, @_)";
-	call_it($f, $o, @_);
-    };
-}
-
-sub compose {			# '.' operator
-    my @funcs = @_;
-    Dparse "using (@funcs)";
-    return sub {
-	no strict 'refs';
-	Drun "compose (@funcs)";
-	foreach (reverse @funcs) {
-	    @_ = call_it($_, @_);
+    if (ref $sym eq $type) {
+	return $sym;
+    }
+    my ($where, $thing);
+    foreach ('Language::FP', pkg()) {
+	my $x = $_.'::'.$sym;
+	if (defined($thing = *{$x}{$type})) {
+	    $where = $x;
+	    last;
 	}
-	@_[0..$#_];
-    };
-}
-
-sub distribute {		# '[...]' list-of-functions
-    my @xs = @_;
-    Dparse "using (@xs)";
-    return sub {
-	no strict 'refs';
-	Drun "distribute (@xs) : (@_)";
-	map { to_array call_it $_, @_ } @xs;
+    }
+    if (wantarray) {
+	($thing, $where);
+    } else {
+	$thing;
     }
 }
 
-sub ifelse {			# 'a -> b ; c' construct
-    my ($if, $then, $else) = @_;
-    Dparse "if $if then $then else $else";
-    return sub {
-	# XXX: having to call this in array context sucks, but is necessary.
-	Drun "if $if then $then else $else";
-	if ((call_it $if, @_)[0]) {
-	    call_it $then, @_;
-	} else {
-	    call_it $else, @_;
-	}
-    };
-}
+######################################################################
+## The parser
 
-sub awhile {			# 'while x y'
-    my ($while, $do) = @_;
-    Dparse "while ($while) $do";
-    return sub {
-	Drun "while ($while) $do -> (@_)";
-	while ((call_it $while, @_)[0]) {
-	    @_ = call_it $do, @_;
-	}
-	@_;
-    }
-}
-
-sub forall {			# '@' operator, i.e. map
-    my $f = shift;
-    Dparse "using $f";
-    return sub {
-	no strict 'refs';
-	Drun "forall $f (@_)";
-	map { to_array call_it $f, as_array $_ } @_;
-    };
-}
-
-sub insert {			# '/' operator, i.e. reduce
-    my $f = shift;
-    Dparse "using $f";
-    return sub {
-	no strict 'refs';
-	Drun "insert $f (@_)";
-	return () unless @_;
-	my $r = $_[0];
-	foreach (@_[1..$#_]) {
-	    $r = (call_it $f, $r, $_)[0];
-	}
-	$r;
-    }
-}
-
-sub constant {			# constant '`' operator
-    my $x = shift;
-    Dparse $x;
-    return sub {
-	Drun "constant $x";
-	as_array $x;
-    };
-}
-
-my %ops;
-sub opfunc {			# symbol table for binary operators
-    use Carp 'confess';
-    confess "$_[0]" unless exists $ops{$_[0]};
-    return $ops{$_[0]};
-}
-
-local $::fp_caller = 'Language::FP';
-sub pkg {			# package in which to bind functions
-    $::fp_caller;
-}
-
-my $P;				# the parser
-sub make_parser {
+my $P = undef;
+sub get_parser {
+    return $P if $P;
     $P = new Parse::RecDescent <<'EOG' or die "Can't create parser!";
 
 {
-	use Regexp::Common;
-	BEGIN {
-		no strict 'refs';
-		foreach my $sym (qw|as_array Dparse Drun opfunc BOTTOM compose
-				    awhile forall ifelse do_bu distribute
-				    insert pkg to_array constant call_it
-				    findsym|) {
-			*{$sym} = *{'Language::FP::'.$sym};
-		}
+use Regexp::Common;
+    BEGIN {
+	no strict 'refs';
+	foreach (qw|term findsym|) {
+	    *{__PACKAGE__.'::'.$_} = \&{'Language::FP::'.$_};
 	}
+    }
 }
 
-
-thing:	  'val' <commit> id_undef '=' application {
-		no strict 'refs';
-		@{pkg().'::'.$item{id_undef}} = as_array $item{application};
-		Dparse "Defined variable $item{id}";
-		$return = 'ok';
-	}
-
-	| 'def' <commit> id_undef '=' termlist {
-		no strict 'refs';
-		*{pkg().'::'.$item{id_undef}} = $item{termlist};
-		Dparse "Defined function $item{id}";
-		$return = 'ok';
-	}
+thing:	  'val' <commit> id_undef '=' application
+		{ $return = term 'val', @item{qw(id_undef application)} }
+	| 'def' <commit> id_undef '=' termlist
+		{ $return = term 'def', @item{qw(id_undef termlist)} }
 	| application
-		{ Dparse "Successful application $item[1]"; $return = $item[1] }
+		{ $return = $item[1];1; }
+	| /\s*/
 	| <error>
 
-application: termlist ':' <commit> data {
-		no strict 'refs';
-		Dparse "application of $item[1]";
-		my @a = as_array $item{data};
-		$return = [ call_it $item[1], @a ];
-	}
+application: termlist ':' <commit> data
+		{ $return = term 'application', @item{qw(termlist data)}; }
 	| data
-		{ $return = $item{data} }
+		{ $return = $item[1]; }
 
 termlist: 'while' <commit> complist termlist
-		{ $return = awhile $item{complist}, $item{termlist} }
+		{ $return = term 'while', @item{qw(complist termlist)} }
  	| complist '->' <commit> complist ';' termlist
- 		{ $return = ifelse @item[1,4,6] }
+		{ $return = term 'if', @item[1,4,6] }
 	| complist 
-		{ $return = $item[1] }
+		{ $return = $item[1];1; }
 	| <error>
 
 complist: <rightop: func '.' func>
-		{ $return = compose @{$item[1]} }
+		{ $return = term 'compose', @{$item[1]} }
 
 func:	  'bu' <commit> func data
-		{ $return = do_bu @item{'func', 'data'} }
+		{ $return = term 'bu', @item{qw(func data)} }
 	| '/' func
-		{ $return = insert $item{func} }
+		{ $return = term 'insert', $item{func} }
 	| '@' <commit> func
-		{ $return = forall $item{func};	}
+		{ $return = term 'forall', $item{func} }
 	| '(' <commit> termlist ')'
 		{ $return = $item{termlist} }
 	| '[' <commit> <rightop: termlist ',' termlist> ']'
-		{ $return = distribute @{$item[3]} }
+		{ $return = term 'distribute', @{$item[3]} }
 	| '`' <commit> data
-		{ $return = constant $item{data} }
+		{ $return = term 'constant', $item{data} }
 	| sfunc
- 		{ $return = $item[1] }
+		{ $return = $item[1];1; }
 	| id
-		{ $return = $item[1] }
+		{ $return = $item[1];1; }
 	| <error>
 
-data:	  '<' <commit> data(s?) '>'
-		{ $return = $item[3]; }
+data:	  datum
+		{ $return = term 'data', $item[1] }
+	| <error>
+
+datum:	  '<' <commit> datum(s?) '>'
+		{ $return = $item[3];1; }
 	| /$RE{num}{real}/o
-		{ $return = $item[1] }
+		{ $return = $item[1];1; }
  	| /$RE{num}{int}/o
-		{ $return = $item[1] }
+		{ $return = $item[1];1; }
 	| /$RE{quoted}/o
-		{ $return = substr($item[1], 1, length($item[1]) - 2) }
+		{ $return = substr($item[1], 1, length($item[1]) - 2);1; }
 	| m{[a-rt-zA-Z_][\w\d]*}
 	  <error?: Undefined variable "$item[1]"> <commit> {
 		no strict 'refs';
+		# XXX: actually interpolate variables during parse.
 		$return = findsym($item[1], 'ARRAY') || undef;
 	}
 	| <error>
 
-sfunc:	  /\d+/ {
-		my $x = $item[1];
- 		$return = sub { $_[$x - 1] };
-	}
+sfunc:	  /\d+/
+		{ $return = term 'sfunc', $item[1]; }
 
 id_undef:  m{[a-zA-Z_][\w\d]*}
-		{ $return = $item[1] }
+		{ $return = term 'id_undef', $item[1]; }
 
 id:	  m{[a-zA-Z_][\w\d]*}
-	  <error?: Undefined function "$item[1]"> <commit>
-		{ $return = findsym($item[1], 'CODE') }
+		{ $return = term 'id', $item[1];1; }
 	| m{([!<>=]=) | [+*/<>-] | ([gln]e) | ([gl]t) | eq}x
-	  <error?: Undefined operator "$item[1]">
-		{ $return = opfunc($item[1]) }
+		{ $return = term 'op', $item[1]; }
 
 EOG
+    $P;
 }
 
 ######################################################################
-## Builtin functions:
+## Builtin functions (for both compilers).
 
 # FP is supposed to be "bottom-preserving".  In other words, once a
 # single operaation fails, it taints all results that depend on it.
 # The only way to recover from this is to explicitly recognize the
 # "bottom" condition using the bottom() test.
 
-sub import {
-    Language::FP->export_to_level(1, @_);
-
-    # XXX: maybe consider autoloading these on demand.
-
-    # Build op-functions.
-    my %make_ops = (
+my %op_guts;
+BEGIN {
+    %op_guts = (
 ## List ops #####
 # first/last element of list
 hd 	=> '@_ ? $_[0] : BOTTOM',
@@ -356,12 +272,139 @@ trans 	=> q{
     @ret;
 },
 );
+}
 
-    while (my ($f, $b) = each %make_ops) {
-	*{$f} = eval qq{ sub { return BOTTOM if bottom(\@_); $b }};
-	die "$f: $@" if $@;
+######################################################################
+## Closure-based "Compiler"
+
+sub defun {			# 'def' X '=' ...
+    my ($name, $val) = @_;
+    no strict 'refs';
+    *{pkg().'::'.$name} = $val;
+    Drun "Defined function $name";
+    'ok';
+}
+
+sub defvar {			# 'val' X '=' ...
+    my ($name, $val) = @_;
+    no strict 'refs';
+    @{pkg().'::'.$name} = as_array $val;
+    Drun "Defined value $name";
+    'ok';
+}
+
+sub do_bu {			# bu (i.e. currying)
+    my ($f, $o) = @_;
+    Dparse "using $f($o, ...)";
+    return sub {
+	no strict 'refs';
+	Drun "bu $f ($o, @_)";
+	call_it($f, $o, @_);
+    };
+}
+
+sub compose {			# '.' operator
+    my @funcs = @_;
+    Dparse "using (@funcs)";
+    return sub {
+	no strict 'refs';
+	Drun "compose (@funcs)";
+	foreach my $f (reverse @funcs) {
+	    @_ = call_it($f, @_);
+	}
+	@_;
+    };
+}
+
+sub distribute {		# '[...]' list-of-functions
+    my @xs = @_;
+    Dparse "using (@xs)";
+    return sub {
+	no strict 'refs';
+	Drun "distribute (@xs) : (@_)";
+	map { to_arrayref call_it $_, @_ } @xs;
     }
+}
 
+sub ifelse {			# 'a -> b ; c' construct
+    my ($if, $then, $else) = @_;
+    Dparse "if $if then $then else $else";
+    return sub {
+	# XXX: having to call this in array context sucks, but is necessary.
+	Drun "if $if then $then else $else";
+	my ($test) = call_it $if, @_;
+	if (bottom($test)) {
+	    BOTTOM;
+	} elsif ($test) {
+	    call_it $then, @_;
+	} else {
+	    call_it $else, @_;
+	}
+    };
+}
+
+sub awhile {			# 'while x y'
+    my ($while, $do) = @_;
+    Dparse "while ($while) $do";
+    return sub {
+	Drun "while ($while) $do -> (@_)";
+	my $test;
+	while (!bottom($test = (call_it $while, @_)[0])) {
+	    if (!$test) {
+		Drun "END while ($while): (@_)";
+		return @_;
+	    }
+	    @_ = call_it $do, @_;
+	}
+	# Bottom.
+	BOTTOM;
+    }
+}
+
+sub forall {			# '@' operator, i.e. map
+    my $f = shift;
+    Dparse "using $f";
+    return sub {
+	no strict 'refs';
+	Drun "forall $f (@_)";
+	map { to_arrayref call_it $f, as_array $_ } @_;
+    };
+}
+
+sub insert {			# '/' operator, i.e. reduce
+    my $f = shift;
+    Dparse "using $f";
+    return sub {
+	no strict 'refs';
+	Drun "insert $f (@_)";
+	return () unless @_;
+	my $r = $_[0];
+	return BOTTOM if bottom($r);
+	foreach (@_[1..$#_]) {
+	    $r = (call_it $f, $r, $_)[0];
+	    return BOTTOM if bottom($r);
+	}
+	$r;
+    }
+}
+
+sub constant {			# constant '`' operator
+    my $x = shift;
+    Dparse $x;
+    return sub {
+	Drun "constant $x";
+	as_array $x;
+    };
+}
+
+sub apply {			# ':' operator
+    my ($func, $args) = @_;
+    return $func->(as_array $args);
+}
+
+my %ops = ();			# symbol table for binary operators
+sub make_binary_ops {
+    return if keys %ops > 0;
     # Build binary operator functions.
     foreach my $f (qw|+ - * / ** == != < > <= >=|) {
 	$ops{$f} = eval qq{sub {
@@ -370,12 +413,283 @@ trans 	=> q{
 		}
 		} || die $@;
     }
-    make_parser;
-    1;
+}
+
+local $::fp_caller = 'Language::FP';
+sub pkg {			# package in which to bind functions
+    $::fp_caller;
+}
+
+my %compile =
+(
+ val => \&defvar,
+ def => \&defun,
+ application => \&apply,
+ while => \&awhile,
+ if => \&ifelse,
+ compose => \&compose,
+ bu => \&do_bu,
+ insert => \&insert,
+ forall => \&forall,
+ distribute => \&distribute,
+ constant => \&constant,
+ sfunc => sub {
+     my $x = $_[0] || die "sfunc($#_): (@_)";
+     sub { $_[$x - 1] }
+ },
+ id => sub {
+     my $ret = findsym($_[0], 'CODE');
+     unless ($ret) {
+	 warn "Undefined function $_[0].";
+	 return \&BOTTOM;
+     }
+     $ret;
+ },
+ data => sub { @_ },
+ id_undef => sub { shift },
+ op => sub {
+     confess "unknown operator '$_[0]'" unless exists $ops{$_[0]};
+     return $ops{$_[0]};
+ }
+);
+
+sub closure_compile {		# internal compiler function
+    my $tree = shift;
+    if (ref $tree ne 'HASH') {
+	return $tree;
+    }
+    my $type = $tree->{type};
+    if (exists $compile{$type}) {
+	my @args = map { closure_compile($_) } @{$tree->{val}};
+	return $compile{$type}->(@args);
+    } else {
+	die "Can't handle $tree (type = $type)";
+    }
+}
+
+sub CLOSURE_compile {		# external compiler function
+    make_binary_ops;
+    closure_compile(@_);
+}
+
+######################################################################
+# The "Big Heinous Eval" compiler.
+
+=for comment
+
+Since Perl's sub calls are slow, I decided to try compiling FP def's
+down to single, heinous Perl functions.  As I suspected, this turns
+out to be much faster than the other implementation, though debugging
+is much more of a challenge.
+
+Each code generating function should return an expression that will
+evaluate to its result in list context, and that has enough parens
+around it to avoid confusing Perl's parser.
+
+The functions should use temporaries where necessary to avoid
+evaluating any of its arguments more than once.  These temporaries
+cannot be references, since the arguments generally won't be real
+arrays, but expressions producing them.
+
+=cut
+
+sub seq($) {			# Turn a sequence into an expression
+    return '(do {
+'.$_[0].'
+})';
+}
+
+my $gen = 0;
+sub gensym {			# yep.
+    if (wantarray) {
+	return map { $_.'_'.++$gen } @_;
+    } else {
+	return $_[0].'_'.++$gen;
+    }
+}
+
+my %bhe_compile =
+(
+ val => sub {
+     my ($name, $val, $rhs) = @_;
+     $name = bhe_compile($name, undef);
+     eval '@'.pkg()."::$name = ".bhe_compile($val, $rhs).";'ok';";
+     if ($@) {
+	 warn "Compilation error in $name: $@" if $@;
+	 BOTTOM;
+     } else {
+	 q{'ok'};
+     }
+ },
+ def => sub {
+     my ($name, $val, $rhs) = @_;
+     my $BODY = bhe_compile($val, $rhs);
+     my $NAME = pkg().'::'.bhe_compile($name, undef);
+     Dparse "\n---\n", $body, "\n---\n";
+     eval "sub $NAME { $BODY }";
+     if ($@) {
+	 warn "Compilation error in $NAME: $@" if $@;
+	 BOTTOM;
+     } else {
+	 q{'ok'};
+     }
+ },
+ application => sub {
+     my ($func, $args, $rhs) = @_;
+     my $arg = bhe_compile($args, undef);
+     return bhe_compile($func, $arg);
+ },
+ while => sub {
+     my ($while, $do, $rhs) = @_;
+     my ($test, $res) = gensym '$WHILE', '@WHILE';
+     my $WHILE = bhe_compile($while, $res);
+     my $DO = bhe_compile($do, $res);
+     seq <<END;
+my $res = $rhs;
+while (my $test = ($WHILE)[0]) {
+	$res = $DO;
+	return BOTTOM if bottom($res);
+    }
+     $res;
+END
+ },
+ if => sub {
+     my ($if, $then, $else, $rhs) = @_;
+     my $x = gensym '@IF';
+     my ($IF, $THEN, $ELSE)
+	 = map { bhe_compile($_, $x) } ($if, $then, $else);
+     seq <<END;
+my $x = $rhs;
+if (($IF)[0]) {
+    $THEN;
+} else {
+    $ELSE;
+}
+END
+ },
+ compose => sub {
+     my $rhs = pop;
+     my @funcs = @_;
+     my $ret = $rhs;
+     while (my $x = pop @funcs) {
+	 $ret = bhe_compile($x, $ret);
+     }
+     $ret;
+ },
+ bu => sub {
+     my ($f, $a, $rhs) = @_;
+     my $A = bhe_compile($a, undef);
+     return bhe_compile($f, "($A, $rhs)");
+ },
+ insert => sub {
+     my ($f, $rhs) = @_;
+     my ($r, $x, $xs) = gensym '$INSERT', '$INSERT', '@INSERT';
+     my $DOIT = bhe_compile($f, "($r, $x)");
+     return seq <<END;
+my $xs = $rhs;
+if ($xs) {
+    my $r = shift $xs;
+    foreach my $x ($xs) {
+	$r = ($DOIT)[0];
+	return Language::FP::BOTTOM if Language::FP::bottom($r);
+    }
+    $r;
+} else {
+    ();				# nothing to insert
+}
+END
+},
+ forall => sub {
+     my ($f, $rhs) = @_;
+     my $v = gensym '@FORALL';
+     my $body = bhe_compile($f, $v);
+     seq <<ENDS;
+map {
+    my $v = Language::FP::as_array(\$_);
+    Language::FP::to_arrayref($body)
+} $rhs
+ENDS
+ },
+ distribute => sub {
+     my $rhs = pop;
+     my $args = gensym '@DISTRIBUTE';
+     my $ret = "my $args = $rhs;\n(";
+     $ret .= join ",\n\t", map {
+	 'Language::FP::to_arrayref('.bhe_compile($_, $args).')'
+     } @_;
+     seq ($ret . ');');
+ },
+ constant => sub { return bhe_compile(shift, undef) },
+ sfunc => sub {
+     my ($x, $rhs) = @_;
+     --$x;			# FP indices are one-based.
+     "(($rhs)[$x])";
+ },
+ id => sub {
+     my ($f, $rhs) = @_;
+     my ($code, $fullname) = findsym($f, 'CODE');
+     unless ($code) {
+	 warn "Undefined function $f.";
+	 return 'return BOTTOM';
+     }
+     unless ($fullname) {
+	 warn "Anonymous sub not supported\n";
+	 return 'return BOTTOM';
+     }
+     return '&{'.$fullname.'}('.$rhs.')';
+ },
+ data => sub {
+     use Data::Dumper;
+     pop;			# get rid of rhs.
+     return 'Language::FP::as_array('.seq(Dumper(to_arrayref @_)).')';
+ },
+ id_undef => sub { shift },
+ op => sub {
+     my ($op, $rhs) = @_;
+     my $res = gensym 'OP';
+     seq "my \@$res = $rhs; \$$res\[0] $op \$$res\[1]";
+ }
+);
+
+sub BHE_compile {		# bootstrap function for BHE
+    my $compiled = bhe_compile(@_, '@_');
+    Dparse "---\n$compiled\n---\n";
+    my @ret = eval $compiled;
+    warn $@ if $@;
+    @ret;
+}
+
+sub bhe_compile {		# Internal BHE compile function
+    my $tree = shift;
+    my $rhs = shift;
+    if (ref $tree ne 'HASH') {
+	# Terminals should never call bhe_compile
+	die;
+    }
+    my $type = $tree->{type};
+    if (exists $compile{$type}) {
+	return $bhe_compile{$type}->(@{$tree->{val}}, $rhs);
+    } else {
+	die "Can't handle $tree (type = $type)";
+    }
 }
 
 ######################################################################
 ## Exportables:
+
+sub import {
+    Language::FP->export_to_level(1, @_);
+
+    # XXX: maybe consider autoloading these?
+
+    # Build op-functions.
+    while (my ($f, $b) = each %op_guts) {
+	*{$f} = eval qq{ sub { return BOTTOM if bottom(\@_); $b }};
+	die "$f: $@" if $@;
+    }
+    1;
+}
+
 sub perl2fp {
     my @ret;
     foreach (@_) {
@@ -396,13 +710,24 @@ sub perl2fp {
 
 sub fp2perl {
     my $str = shift;
-    return to_array($P->data($str));
+    return to_arrayref($P->data($str));
 }
 
 sub fp_eval {
     local $::fp_caller = caller;
+    my $p = get_parser;
     if (@_ == 1) {
-	return $P->thing(shift);
+	use Data::Dumper;
+	my $parsed = $P->thing(shift);
+	unless ($parsed) {
+	    warn "Parse error";
+	    return undef;
+	}
+ 	if ($::FP_DEBUG =~ /C/) {
+	    return [CLOSURE_compile($parsed)];
+ 	} else {
+	    return [BHE_compile($parsed)];
+ 	}
     }
 
     my %o = @_;
@@ -510,7 +835,7 @@ C<distl> in Perl, one could write
 
 You will experience unexpected behavior when programming in FP.  Some
 of it may even be your fault.  When this occurs, setting the global
-variable C<$::DEBUG> to a string containing one or more of the
+variable C<$::FP_DEBUG> to a string containing one or more of the
 following characters can help:
 
 =over
@@ -521,6 +846,8 @@ following characters can help:
 
 =item 'b' -- Make FP errors ("bottom") fatal
 
+=item 'C' -- Use the slower, closure-based evaluator.
+
 =back
 
 =head1 EXPORTS
@@ -532,7 +859,7 @@ command-line convenience.
 
 Documentation -- a lot more needs to be explained a lot better.
 
-Testing -- only lightly tested, though it can handle totient.
+Testing -- getting better, but still needs work.
 
 Maybe make it more "OO" -- not that important.
 
@@ -568,4 +895,4 @@ argument.  This may be the longest FP program ever written.
 =cut
 
 # one-liner version of the above, for cut-and-paste:
-# def totient = /+ . @((== . [1, `1] -> `1 ; `0) . (while (> . [2, `0]) (< -> reverse ; id) . [2, -])) . distl . [id, iota]
+# def totient = /+ . @((== . [1, `1] -> `1 ; `0) . (while > . [2, `0] (< -> reverse ; id) . [2, -])) . distl . [id, iota]
